#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include <signal.h>
#include <sys/wait.h>
#include <tcl.h>
#include <tk.h> 
#include "tkExtWin.h"

static Tk_ConfigSpec 
configSpecs[] = {

    {TK_CONFIG_BORDER, "-background", "background", "Background",
        "#cdb79e", Tk_Offset(ExtWin, bgBorder), TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
        "white", Tk_Offset(ExtWin, bgBorder), TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
        (char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
        (char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
        "5", Tk_Offset(ExtWin, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
        (char *) NULL, 0, 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
        "sunken", Tk_Offset(ExtWin, relief), 0},
    {TK_CONFIG_INT,"-width","width","Width","400",Tk_Offset(ExtWin,width),
        0},
    {TK_CONFIG_INT,"-height","height","Height","300",Tk_Offset(ExtWin,height),
        0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
        (char *) NULL, 0, 0}
};

/*
 *--------------------------------------------------------------
 *
 * ExtWinCmd --
 *
 *	This procedure is invoked to process the "ExternalWin" Tcl
 *	command.  It creates a new "ExternalWin" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */
int
ExtWinCmd(ClientData clientData,Tcl_Interp * interp,int argc,char ** argv){

    Tk_Window mainw = (Tk_Window) clientData;
    ExtWin *extWinPtr;
    Tk_Window tkwin;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_CreateWindowFromPath(interp, mainw, argv[1], (char *) NULL);

    if (tkwin == NULL) return TCL_ERROR;

    Tk_SetClass(tkwin, "ExtWin");

    /*
     * Allocate and initialize the widget record.
     */

    extWinPtr = (ExtWin *) ckalloc(sizeof(ExtWin));
    extWinPtr->main = mainw;
    extWinPtr->tkwin = tkwin;
    extWinPtr->display = Tk_Display(tkwin);
    extWinPtr->win = Tk_WindowId(tkwin);
    extWinPtr->screen = Tk_Screen(tkwin);
    extWinPtr->interp = interp;
    extWinPtr->width = 260;
    extWinPtr->height = 260;
    extWinPtr->borderWidth = 0;
    extWinPtr->bgBorder = NULL;
//    extWinPtr->fgBorder = NULL;
    extWinPtr->relief = TK_RELIEF_FLAT;
    extWinPtr->gc = None;
    extWinPtr->childPid = 0;
    extWinPtr->updatePending = 0;
    extWinPtr->pixmap = (Pixmap) NULL;
    extWinPtr->pipe_to_rem[0] =  extWinPtr->pipe_to_rem[1] = -1;
    extWinPtr->pipe_from_rem[0] =  extWinPtr->pipe_from_rem[1] = -1;

    Tk_CreateEventHandler(extWinPtr->tkwin, ExposureMask|StructureNotifyMask,
	    ExtWinEventProc, (ClientData) extWinPtr);
    Tcl_CreateCommand(interp, Tk_PathName(extWinPtr->tkwin), ExtWinWidgetCmd,
	    (ClientData) extWinPtr, (void (*)()) NULL);
    if (ExtWinConfigure(interp, extWinPtr, argc-2, argv+2, 0) != TCL_OK) {
	Tk_DestroyWindow(extWinPtr->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(extWinPtr->tkwin);
    return TCL_OK;
}
/*
 * ----------------------------------------------------------------------
 *
 *	ExtWin init --
 *
 *		Initialize the Video-Widget
 *
 * ----------------------------------------------------------------------
 */
int ExtWin_Init(Tcl_Interp *interp) {
        Tcl_CreateCommand ( interp,"extwin",ExtWinCmd,(ClientData) Tk_MainWindow
                (interp),(Tcl_CmdDeleteProc *) NULL);
        return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ExtWinWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
ExtWinWidgetCmd(ClientData clientData,Tcl_Interp* interp,int argc,char ** argv)
{
    ExtWin *extWinPtr = (ExtWin *) clientData;
    int result = TCL_OK;
    int length;
    int do_update =0;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) extWinPtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	do_update = 1;
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, extWinPtr->tkwin, configSpecs,
		    (char *) extWinPtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, extWinPtr->tkwin, configSpecs,
		    (char *) extWinPtr, argv[2], 0);
	} else {
	    result = ExtWinConfigure(interp, extWinPtr, argc-2, argv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    }
    else if (( c =='w') && (strncmp(argv[1],"window",length) == 0)) {
	if(argc != 2 ) { 
	 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " window \"", (char *) NULL);
		    goto error;
	}
	sprintf(interp->result, "%ld",Tk_WindowId(extWinPtr->tkwin));
    } 
    else if (( c == 'd' ) && (strncmp(argv[1],"display",length) == 0)) {
	if(argc != 2 ) {
		    Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " window \"", (char *) NULL);
                    goto error;
        }
	sprintf(interp->result,"%s",XDisplayName(""));
    }

/*
 * -------------------------------------------------------------------
 *  Starten externer Programme 
 * -------------------------------------------------------------------
 */

    else if (( c == 's') && (strncmp(argv[1],"start",length) == 0)) {
	if( argc < 3 ){
                   Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], "start programmname ?args?\"", (char *) NULL);
                    goto error;
	}

/*
 * ----------------------------------------------------------------------
 *	Create Pipes for communication with the external process
 * ----------------------------------------------------------------------
 */

	if(pipe(extWinPtr->pipe_to_rem) != 0){
		Tcl_AppendResult(interp,"pipe open error",(char *) NULL);
		goto error;
	}
	if(pipe(extWinPtr->pipe_from_rem) != 0){
		close(extWinPtr->pipe_to_rem[0] );
		close(extWinPtr->pipe_to_rem[1] );
		Tcl_AppendResult(interp,"pipe open error",(char *) NULL);
                goto error;
	}
	fcntl(extWinPtr->pipe_from_rem[0],F_SETFL,O_NONBLOCK,O_SYNC);
	if(!(extWinPtr->childPid = fork()) ){
	
		close(0);	  		 	/* stdin */
                dup(extWinPtr->pipe_to_rem[0]);	
                close(extWinPtr->pipe_to_rem[0]);
                close(extWinPtr->pipe_to_rem[1]);
		close(1);				/* stdout  */
                dup(extWinPtr->pipe_from_rem[1]);
                close(extWinPtr->pipe_from_rem[0]);
                close(extWinPtr->pipe_from_rem[1]);
	
                /*execve(argv[2],argv+2,NULL);*/
		execv(argv[2],argv+2);
	        fprintf(stderr,"can't execute %s\n",argv[2]);	
                exit (0);
        }
	close(extWinPtr->pipe_to_rem[0]);
	extWinPtr->pipe_to_rem[0]=-1;
	close(extWinPtr->pipe_from_rem[1]);
	extWinPtr->pipe_from_rem[1]=-1;
	sprintf(interp->result,"%s","");
    }

/* ------------------------------------------------------------------------- 
 *
 *     Send commands to stdin of the extenal process
 *     There is no test for success
 *
 *-------------------------------------------------------------------------
 */

    else if (( c == 's') && (strncmp(argv[1],"send",length) == 0)) { 
	int i;
	 if( argc < 3 ){
                   Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], "send <string>" ,(char *) NULL);
                    goto error;
        }
	for(i=2;i<argc;i++){
		write(extWinPtr->pipe_to_rem[1],argv[i],strlen(argv[i]));
	}
	write(extWinPtr->pipe_to_rem[1],"\n",1);
        	
    } 
/* ------------------------------------------------------------------------------
 *
 *     		Loeschen des Puffers und lesen aus der Pipe
 *
 * ------------------------------------------------------------------------------
 */
    else if (( c == 'c' ) && (strncmp(argv[1],"clear",length)== 0 )) {
	char c;
	while(read(extWinPtr->pipe_from_rem[0],&c,1) > 0);
    }
    else if (( c == 'g' ) && (strncmp(argv[1],"get",length) == 0 )){
	int i,timeout=40,strsize = 2048;
	char c,string[2048] ;
	char *str = string; 
	if( argc < 2 ){
                   Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], "get" ,(char *) NULL);
                    goto error;
        }
	while( ((i=read(extWinPtr->pipe_from_rem[0],&c,1))<=0) && (timeout > 0) ){
	    timeout--;
	    usleep(100000);
	}
	if(i<=0){
		 Tcl_AppendResult(interp, "timeout waiting for answer of child ",
                    (char *)NULL);
                    goto error;
	}
	*string = c; 
	while((i>0) && (strsize > 4 )){
		i=read(extWinPtr->pipe_from_rem[0],++str,1);
		if((i>0) && (*str == '\n'))str--;
		else strsize --;
	}
	*str = 0;  
	Tcl_AppendResult(interp,string,NULL);
   }


    else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, window, display, send, or start", (char *) NULL);
	goto error;
    }
    if (!extWinPtr->updatePending && do_update ) {
	Tk_DoWhenIdle(ExtWinDisplay, (ClientData) extWinPtr);
	extWinPtr->updatePending = 1;
    }
    Tk_Release((ClientData) extWinPtr);
    return result;

    error:
    Tk_Release((ClientData) extWinPtr);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * ExtWinConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a ExtWin widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for extWinPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ExtWinConfigure(Tcl_Interp *interp,ExtWin *extWinPtr,
		int argc,char **argv, int flags){

    if (Tk_ConfigureWidget(interp, extWinPtr->tkwin, configSpecs,
	    argc, argv, (char *) extWinPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Set the background for the window and create a graphics context
     * for use during redisplay.
     */

    Tk_SetWindowBackground(extWinPtr->tkwin,
	    Tk_3DBorderColor(extWinPtr->bgBorder)->pixel);

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */

    Tk_GeometryRequest(extWinPtr->tkwin, extWinPtr->width, extWinPtr->height);
    Tk_SetInternalBorder(extWinPtr->tkwin, extWinPtr->borderWidth);
    if (!extWinPtr->updatePending) {
	Tk_DoWhenIdle(ExtWinDisplay, (ClientData) extWinPtr);
	extWinPtr->updatePending = 1;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ExtWinEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on ExtWins.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
ExtWinEventProc(ClientData clientData,XEvent *eventPtr)
{
    ExtWin *extWinPtr = (ExtWin *) clientData;

    if (eventPtr->type == Expose) {
	if (!extWinPtr->updatePending) {
	    Tk_DoWhenIdle(ExtWinDisplay, (ClientData) extWinPtr);
	    extWinPtr->updatePending = 1;
	}
    } else if (eventPtr->type == ConfigureNotify) {
	if (!extWinPtr->updatePending) {
	    Tk_DoWhenIdle(ExtWinDisplay, (ClientData) extWinPtr);
	    extWinPtr->updatePending = 1;
	}
    } else if (eventPtr->type == DestroyNotify) {
	if(extWinPtr->pipe_from_rem[0]>=0) {
		close(extWinPtr->pipe_from_rem[0]);
		extWinPtr->pipe_from_rem[0]=-1;
	}
	if(extWinPtr->pipe_to_rem[1]>=0) {
		close(extWinPtr->pipe_to_rem[1]);
		extWinPtr->pipe_to_rem[1]=-1;
	}
	Tcl_DeleteCommand(extWinPtr->interp, Tk_PathName(extWinPtr->tkwin));
	extWinPtr->tkwin = NULL;
	if (extWinPtr->updatePending) {
	    Tk_CancelIdleCall(ExtWinDisplay, (ClientData) extWinPtr);
	}
	Tk_EventuallyFree((ClientData) extWinPtr, ExtWinDestroy);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ExtWinDisplay --
 *
 *	This procedure redraws the contents of a ExtWin window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
ExtWinDisplay(ClientData clientData)
{
    ExtWin *extWinPtr = (ExtWin *) clientData;
    Tk_Window tkwin = extWinPtr->tkwin;
    Drawable d;

    extWinPtr->updatePending = 0;
    if (!Tk_IsMapped(tkwin)) {
	return;
    }
     {
	Window list[]={Tk_WindowId(extWinPtr->tkwin),Tk_WindowId(extWinPtr->main)};

        XSetWMColormapWindows(extWinPtr->display,Tk_WindowId(extWinPtr->main),
                        list,2);
    }
     /* 
      * Redraw the widget's background and border.
      */

    d = Tk_WindowId(tkwin);
    Tk_Fill3DRectangle(tkwin, d, extWinPtr->bgBorder,
	    0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
	    extWinPtr->borderWidth, extWinPtr->relief); 

    XSync(extWinPtr->display,False); // If we don't wait the 3d-Rectangle 
				     // is probably drawn after the expose is handled.

    if(extWinPtr->pipe_to_rem[1]) {
    	write(extWinPtr->pipe_to_rem[1],"expose",strlen("expose"));
    	write(extWinPtr->pipe_to_rem[1],"\n",1);
    }
 


}

/*
 *----------------------------------------------------------------------
 *
 * ExtWinDestroy --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a ExtWin at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the ExtWin is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
ExtWinDestroy(char *blockPtr)
{
    ExtWin *extWinPtr = (ExtWin *) blockPtr;

    Tk_FreeOptions(configSpecs, (char *) extWinPtr, extWinPtr->display, 0);
    if (extWinPtr->gc != None) {
	Tk_FreeGC(extWinPtr->display, extWinPtr->gc);
    }
    if(extWinPtr->childPid) 
		kill(extWinPtr->childPid,15);
    if(waitpid(extWinPtr->childPid,NULL,0)!=extWinPtr->childPid)
	fprintf(stderr,"Unable to kill child process.\n");
    ckfree((char *) extWinPtr);
	
}

